@@ -1,9 +1,9 @@
-use byteorder::{BigEndian, ReadBytesExt};
-use error::{Error, Result};
+use crate::{read_u16_from_be, read_u8};
+use error::{Error, Result, UnsupportedFeature};
 use huffman::{HuffmanTable, HuffmanTableClass};
 use marker::Marker;
 use marker::Marker::*;
-use std::io::Read;
+use std::io::{self, Read};
 use std::ops::Range;
 
 #[derive(Clone, Copy, Debug, PartialEq)]
@@ -34,6 +34,7 @@ pub struct FrameInfo {
     pub precision: u8,
 
     pub image_size: Dimensions,
+    pub output_size: Dimensions,
     pub mcu_size: Dimensions,
     pub components: Vec<Component>,
 }
@@ -58,6 +59,8 @@ pub struct Component {
 
     pub quantization_table_index: usize,
 
+    pub dct_scale: usize,
+
     pub size: Dimensions,
     pub block_size: Dimensions,
 }
@@ -67,6 +70,7 @@ pub enum AppData {
     Adobe(AdobeColorTransform),
     Jfif,
     Avi1,
+    Icc(IccChunk),
 }
 
 // http://www.sno.phy.queensu.ca/~phil/exiftool/TagNames/JPEG.html#Adobe
@@ -78,14 +82,37 @@ pub enum AdobeColorTransform {
     // YCbCrK
     YCCK,
 }
+#[derive(Debug)]
+pub struct IccChunk {
+    pub num_markers: u8,
+    pub seq_no: u8,
+    pub data: Vec<u8>,
+}
+
+impl FrameInfo {
+    pub(crate) fn update_idct_size(&mut self, idct_size: usize) -> Result<()> {
+        for component in &mut self.components {
+            component.dct_scale = idct_size;
+        }
+
+        update_component_sizes(self.image_size, &mut self.components)?;
+
+        self.output_size = Dimensions {
+            width: (self.image_size.width as f32 * idct_size as f32 / 8.0).ceil() as u16,
+            height: (self.image_size.height as f32 * idct_size as f32 / 8.0).ceil() as u16
+        };
+
+        Ok(())
+    }
+}
 
 fn read_length<R: Read>(reader: &mut R, marker: Marker) -> Result<usize> {
     assert!(marker.has_length());
 
     // length is including itself.
-    let length = reader.read_u16::<BigEndian>()? as usize;
+    let length = usize::from(read_u16_from_be(reader)?);
 
-    if length <= 2 {
+    if length < 2 {
         return Err(Error::Format(format!("encountered {:?} with invalid length {}", marker, length)));
     }
 
@@ -93,10 +120,14 @@ fn read_length<R: Read>(reader: &mut R, marker: Marker) -> Result<usize> {
 }
 
 fn skip_bytes<R: Read>(reader: &mut R, length: usize) -> Result<()> {
-    let mut buffer = vec![0u8; length];
-    reader.read_exact(&mut buffer)?;
-
-    Ok(())
+    let length = length as u64;
+    let to_skip = &mut reader.by_ref().take(length);
+    let copied = io::copy(to_skip, &mut io::sink())?;
+    if copied < length {
+        Err(Error::Io(io::ErrorKind::UnexpectedEof.into()))
+    } else {
+        Ok(())
+    }
 }
 
 // Section B.2.2
@@ -109,8 +140,8 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
 
     let is_baseline = marker == SOF(0);
     let is_differential = match marker {
-        SOF(0 ... 3) | SOF(9 ... 11)  => false,
-        SOF(5 ... 7) | SOF(13 ... 15) => true,
+        SOF(0 ..= 3) | SOF(9 ..= 11)  => false,
+        SOF(5 ..= 7) | SOF(13 ..= 15) => true,
         _ => panic!(),
     };
     let coding_process = match marker {
@@ -120,12 +151,12 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
         _ => panic!(),
     };
     let entropy_coding = match marker {
-        SOF(0 ... 3) | SOF(5 ... 7)     => EntropyCoding::Huffman,
-        SOF(9 ... 11) | SOF(13 ... 15)  => EntropyCoding::Arithmetic,
+        SOF(0 ..= 3) | SOF(5 ..= 7)     => EntropyCoding::Huffman,
+        SOF(9 ..= 11) | SOF(13 ..= 15)  => EntropyCoding::Arithmetic,
         _ => panic!(),
     };
 
-    let precision = reader.read_u8()?;
+    let precision = read_u8(reader)?;
 
     match precision {
         8 => {},
@@ -141,18 +172,21 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
         },
     }
 
-    let height = reader.read_u16::<BigEndian>()?;
-    let width = reader.read_u16::<BigEndian>()?;
+    let height = read_u16_from_be(reader)?;
+    let width = read_u16_from_be(reader)?;
 
     // height:
     // "Value 0 indicates that the number of lines shall be defined by the DNL marker and
     //     parameters at the end of the first scan (see B.2.5)."
+    if height == 0 {
+        return Err(Error::Unsupported(UnsupportedFeature::DNL));
+    }
 
     if width == 0 {
         return Err(Error::Format("zero width in frame header".to_owned()));
     }
 
-    let component_count = reader.read_u8()?;
+    let component_count = read_u8(reader)?;
 
     if component_count == 0 {
         return Err(Error::Format("zero component count in frame header".to_owned()));
@@ -168,14 +202,14 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
     let mut components: Vec<Component> = Vec::with_capacity(component_count as usize);
 
     for _ in 0 .. component_count {
-        let identifier = reader.read_u8()?;
+        let identifier = read_u8(reader)?;
 
         // Each component's identifier must be unique.
         if components.iter().any(|c| c.identifier == identifier) {
             return Err(Error::Format(format!("duplicate frame component identifier {}", identifier)));
         }
 
-        let byte = reader.read_u8()?;
+        let byte = read_u8(reader)?;
         let horizontal_sampling_factor = byte >> 4;
         let vertical_sampling_factor = byte & 0x0f;
 
@@ -186,7 +220,7 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
             return Err(Error::Format(format!("invalid vertical sampling factor {}", vertical_sampling_factor)));
         }
 
-        let quantization_table_index = reader.read_u8()?;
+        let quantization_table_index = read_u8(reader)?;
 
         if quantization_table_index > 3 || (coding_process == CodingProcess::Lossless && quantization_table_index != 0) {
             return Err(Error::Format(format!("invalid quantization table index {}", quantization_table_index)));
@@ -197,25 +231,13 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
             horizontal_sampling_factor: horizontal_sampling_factor,
             vertical_sampling_factor: vertical_sampling_factor,
             quantization_table_index: quantization_table_index as usize,
+            dct_scale: 8,
             size: Dimensions {width: 0, height: 0},
             block_size: Dimensions {width: 0, height: 0},
         });
     }
 
-    let h_max = components.iter().map(|c| c.horizontal_sampling_factor).max().unwrap();
-    let v_max = components.iter().map(|c| c.vertical_sampling_factor).max().unwrap();
-    let mcu_size = Dimensions {
-        width: (width as f32 / (h_max as f32 * 8.0)).ceil() as u16,
-        height: (height as f32 / (v_max as f32 * 8.0)).ceil() as u16,
-    };
-
-    for component in &mut components {
-        component.size.width = (width as f32 * (component.horizontal_sampling_factor as f32 / h_max as f32)).ceil() as u16;
-        component.size.height = (height as f32 * (component.vertical_sampling_factor as f32 / v_max as f32)).ceil() as u16;
-
-        component.block_size.width = mcu_size.width * component.horizontal_sampling_factor as u16;
-        component.block_size.height = mcu_size.height * component.vertical_sampling_factor as u16;
-    }
+    let mcu_size = update_component_sizes(Dimensions { width, height }, &mut components)?;
 
     Ok(FrameInfo {
         is_baseline: is_baseline,
@@ -223,23 +245,77 @@ pub fn parse_sof<R: Read>(reader: &mut R, marker: Marker) -> Result<FrameInfo> {
         coding_process: coding_process,
         entropy_coding: entropy_coding,
         precision: precision,
-        image_size: Dimensions {width: width, height: height},
-        mcu_size: mcu_size,
+        image_size: Dimensions { width, height },
+        output_size: Dimensions { width, height },
+        mcu_size,
         components: components,
     })
 }
 
+/// Returns ceil(x/y), requires x>0
+fn ceil_div(x: u32, y: u32) -> Result<u16> {
+    if x == 0 || y == 0 {
+        // TODO Determine how this error is reached. Can we validate input
+        // earlier and error out then?
+        return Err(Error::Format("invalid dimensions".to_owned()));
+    }
+    Ok((1 + ((x - 1) / y)) as u16)
+}
+
+fn update_component_sizes(size: Dimensions, components: &mut [Component]) -> Result<Dimensions> {
+    let h_max = components.iter().map(|c| c.horizontal_sampling_factor).max().unwrap() as u32;
+    let v_max = components.iter().map(|c| c.vertical_sampling_factor).max().unwrap() as u32;
+
+    let mcu_size = Dimensions {
+        width: ceil_div(size.width as u32, h_max * 8)?,
+        height: ceil_div(size.height as u32, v_max * 8)?,
+    };
+
+    for component in components {
+        component.size.width = ceil_div(size.width as u32 * component.horizontal_sampling_factor as u32 * component.dct_scale as u32, h_max * 8)?;
+        component.size.height = ceil_div(size.height as u32 * component.vertical_sampling_factor as u32 * component.dct_scale as u32, v_max * 8)?;
+
+        component.block_size.width = mcu_size.width * component.horizontal_sampling_factor as u16;
+        component.block_size.height = mcu_size.height * component.vertical_sampling_factor as u16;
+    }
+
+    Ok(mcu_size)
+}
+
+#[test]
+fn test_update_component_sizes() {
+    let mut components = [Component {
+        identifier: 1,
+        horizontal_sampling_factor: 2,
+        vertical_sampling_factor: 2,
+        quantization_table_index: 0,
+        dct_scale: 8,
+        size: Dimensions { width: 0, height: 0 },
+        block_size: Dimensions { width: 0, height: 0 },
+    }];
+    let mcu = update_component_sizes(
+        Dimensions { width: 800, height: 280 },
+        &mut components).unwrap();
+    assert_eq!(mcu, Dimensions { width: 50, height: 18 });
+    assert_eq!(components[0].block_size, Dimensions { width: 100, height: 36 });
+    assert_eq!(components[0].size, Dimensions { width: 800, height: 280 });
+}
+
 // Section B.2.3
 pub fn parse_sos<R: Read>(reader: &mut R, frame: &FrameInfo) -> Result<ScanInfo> {
     let length = read_length(reader, SOS)?;
-    let component_count = reader.read_u8()?;
+    if 0 == length {
+        return Err(Error::Format("zero length in SOS".to_owned()));
+    }
+
+    let component_count = read_u8(reader)?;
 
     if component_count == 0 || component_count > 4 {
         return Err(Error::Format(format!("invalid component count {} in scan header", component_count)));
     }
 
     if length != 4 + 2 * component_count as usize {
-        return Err(Error::Format("invalid length in SOF".to_owned()));
+        return Err(Error::Format("invalid length in SOS".to_owned()));
     }
 
     let mut component_indices = Vec::with_capacity(component_count as usize);
@@ -247,7 +323,7 @@ pub fn parse_sos<R: Read>(reader: &mut R, frame: &FrameInfo) -> Result<ScanInfo>
     let mut ac_table_indices = Vec::with_capacity(component_count as usize);
 
     for _ in 0 .. component_count {
-        let identifier = reader.read_u8()?;
+        let identifier = read_u8(reader)?;
 
         let component_index = match frame.components.iter().position(|c| c.identifier == identifier) {
             Some(value) => value,
@@ -264,7 +340,7 @@ pub fn parse_sos<R: Read>(reader: &mut R, frame: &FrameInfo) -> Result<ScanInfo>
             return Err(Error::Format("the scan component order does not follow the order in the frame header".to_owned()));
         }
 
-        let byte = reader.read_u8()?;
+        let byte = read_u8(reader)?;
         let dc_table_index = byte >> 4;
         let ac_table_index = byte & 0x0f;
 
@@ -288,10 +364,10 @@ pub fn parse_sos<R: Read>(reader: &mut R, frame: &FrameInfo) -> Result<ScanInfo>
         return Err(Error::Format("scan with more than one component and more than 10 blocks per MCU".to_owned()));
     }
 
-    let spectral_selection_start = try!(reader.read_u8());
-    let spectral_selection_end = try!(reader.read_u8());
+    let spectral_selection_start = read_u8(reader)?;
+    let spectral_selection_end = read_u8(reader)?;
 
-    let byte = reader.read_u8()?;
+    let byte = read_u8(reader)?;
     let successive_approximation_high = byte >> 4;
     let successive_approximation_low = byte & 0x0f;
 
@@ -344,7 +420,7 @@ pub fn parse_dqt<R: Read>(reader: &mut R) -> Result<[Option<[u16; 64]>; 4]> {
 
     // Each DQT segment may contain multiple quantization tables.
     while length > 0 {
-        let byte = reader.read_u8()?;
+        let byte = read_u8(reader)?;
         let precision = (byte >> 4) as usize;
         let index = (byte & 0x0f) as usize;
 
@@ -368,10 +444,10 @@ pub fn parse_dqt<R: Read>(reader: &mut R) -> Result<[Option<[u16; 64]>; 4]> {
 
         let mut table = [0u16; 64];
 
-        for i in 0 .. 64 {
-            table[i] = match precision {
-                0 => reader.read_u8()? as u16,
-                1 => reader.read_u16::<BigEndian>()?,
+        for item in table.iter_mut() {
+            *item = match precision {
+                0 => u16::from(read_u8(reader)?),
+                1 => read_u16_from_be(reader)?,
                 _ => unreachable!(),
             };
         }
@@ -395,7 +471,7 @@ pub fn parse_dht<R: Read>(reader: &mut R, is_baseline: Option<bool>) -> Result<(
 
     // Each DHT segment may contain multiple huffman tables.
     while length > 17 {
-        let byte = reader.read_u8()?;
+        let byte = read_u8(reader)?;
         let class = byte >> 4;
         let index = (byte & 0x0f) as usize;
 
@@ -451,7 +527,7 @@ pub fn parse_dri<R: Read>(reader: &mut R) -> Result<u16> {
         return Err(Error::Format("DRI with invalid length".to_owned()));
     }
 
-    Ok(reader.read_u16::<BigEndian>()?)
+    Ok(read_u16_from_be(reader)?)
 }
 
 // Section B.2.4.5
@@ -485,7 +561,27 @@ pub fn parse_app<R: Read>(reader: &mut R, marker: Marker) -> Result<Option<AppDa
                     result = Some(AppData::Avi1);
                 }
             }
-        },
+        }
+        APP(2) => {
+            if length > 14 {
+                let mut buffer = [0u8; 14];
+                reader.read_exact(&mut buffer)?;
+                bytes_read = buffer.len();
+
+                // http://www.color.org/ICC_Minor_Revision_for_Web.pdf
+                // B.4 Embedding ICC profiles in JFIF files
+                if &buffer[0..12] == b"ICC_PROFILE\0" {
+                    let mut data = vec![0; length - bytes_read];
+                    reader.read_exact(&mut data)?;
+                    bytes_read += data.len();
+                    result = Some(AppData::Icc(IccChunk {
+                        seq_no: buffer[12],
+                        num_markers: buffer[13],
+                        data,
+                    }));
+                }
+            }
+        }
         APP(14) => {
             if length >= 12 {
                 let mut buffer = [0u8; 12];
